@@ -124,6 +124,8 @@
 #include <linux/rbtree.h>
 #include <linux/cpu.h>              /* CPU hotplug support              */
 
+#include <linux/pm_runtime.h>       /* pm_runtime_*                     */
+
 #include <asm/div64.h>              /* do_div()                         */
 #if defined(NV_ASM_SYSTEM_H_PRESENT)
 #include <asm/system.h>             /* cli, sli, save_flags             */
@@ -173,14 +175,8 @@ static inline uid_t __kuid_val(kuid_t uid)
 #include <linux/vgaarb.h>
 #endif
 
-#if defined(NV_VM_INSERT_PAGE_PRESENT)
 #include <linux/pagemap.h>
-#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 16, 0)
 #include <linux/dma-mapping.h>
-#else
-#include <linux/dma-direct.h>
-#endif
-#endif
 
 #if defined(CONFIG_SWIOTLB) && defined(NVCPU_AARCH64)
 #include <linux/swiotlb.h>
@@ -240,35 +236,14 @@ NV_STATUS nvos_forward_error_to_cray(struct pci_dev *, NvU32,
 
 #if defined(NV_SET_MEMORY_UC_PRESENT)
 #undef NV_SET_PAGES_UC_PRESENT
-#undef NV_CHANGE_PAGE_ATTR_PRESENT
-#elif defined(NV_SET_PAGES_UC_PRESENT)
-#undef NV_CHANGE_PAGE_ATTR_PRESENT
 #endif
 
 #if !defined(NVCPU_FAMILY_ARM) && !defined(NVCPU_PPC64LE)
-#if !defined(NV_SET_MEMORY_UC_PRESENT) && !defined(NV_SET_PAGES_UC_PRESENT) && \
-  !defined(NV_CHANGE_PAGE_ATTR_PRESENT)
+#if !defined(NV_SET_MEMORY_UC_PRESENT) && !defined(NV_SET_PAGES_UC_PRESENT)
 #error "This driver requires the ability to change memory types!"
 #endif
 #endif
 
-/*
- * Using change_page_attr() on early Linux/x86-64 2.6 kernels may
- * result in a BUG() being triggered. The underlying problem
- * actually exists on multiple architectures and kernels, but only
- * the above check for the condition and trigger a BUG().
- *
- * Note that this is a due to a bug in the Linux kernel, not an
- * NVIDIA driver bug.
- *
- * We therefore need to determine at runtime if change_page_attr()
- * can be used safely on these kernels.
- */
-#if defined(NV_CHANGE_PAGE_ATTR_PRESENT) && defined(NVCPU_X86_64) && \
-  (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 11))
-#define NV_CHANGE_PAGE_ATTR_BUG_PRESENT
-#endif
-
 /*
  * Traditionally, CONFIG_XEN indicated that the target kernel was
  * built exclusively for use under a Xen hypervisor, requiring
@@ -291,8 +266,6 @@ NV_STATUS nvos_forward_error_to_cray(struct pci_dev *, NvU32,
 #define NV_XEN_SUPPORT_FULLY_VIRTUALIZED_KERNEL
 #endif
 
-#include "nv-procfs.h"
-
 #ifdef CONFIG_KDB
 #include <linux/kdb.h>
 #include <asm/kdb.h>
@@ -329,29 +302,13 @@ extern int nv_pat_mode;
 
 #if defined(CONFIG_ACPI)
 #include <linux/acpi.h>
-#if defined(NV_ACPI_DEVICE_OPS_HAS_MATCH) || defined(ACPI_VIDEO_HID)
 #define NV_LINUX_ACPI_EVENTS_SUPPORTED 1
 #endif
-#endif
 
 #if (defined(CONFIG_IPMI_HANDLER) || defined(CONFIG_IPMI_HANDLER_MODULE))
 #include <linux/ipmi.h>
 #endif
 
-#if defined(NV_PCI_DMA_MAPPING_ERROR_PRESENT)
-#if (NV_PCI_DMA_MAPPING_ERROR_ARGUMENT_COUNT == 2)
-#define NV_PCI_DMA_MAPPING_ERROR(dev, addr) \
-    pci_dma_mapping_error(dev, addr)
-#elif (NV_PCI_DMA_MAPPING_ERROR_ARGUMENT_COUNT == 1)
-#define NV_PCI_DMA_MAPPING_ERROR(dev, addr) \
-    pci_dma_mapping_error(addr)
-#else
-#error "NV_PCI_DMA_MAPPING_ERROR_ARGUMENT_COUNT value unrecognized!"
-#endif
-#elif defined(NV_VM_INSERT_PAGE_PRESENT)
-#error "NV_PCI_DMA_MAPPING_ERROR() undefined!"
-#endif
-
 #if defined(NV_LINUX_ACPI_EVENTS_SUPPORTED)
 #if (NV_ACPI_WALK_NAMESPACE_ARGUMENT_COUNT == 6)
 #define NV_ACPI_WALK_NAMESPACE(type, args...) acpi_walk_namespace(type, args)
@@ -479,7 +436,7 @@ extern int nv_pat_mode;
         __cpuc_flush_kern_all();
     }
 #define CACHE_FLUSH()            nv_flush_cache_cpu(NULL)
-#define CACHE_FLUSH_ALL()        NV_ON_EACH_CPU(nv_flush_cache_cpu, NULL)
+#define CACHE_FLUSH_ALL()        on_each_cpu(nv_flush_cache_cpu, NULL, 1)
 #if defined(NV_OUTER_FLUSH_ALL_PRESENT)
 #define OUTER_FLUSH_ALL()               outer_flush_all()
 #endif
@@ -493,7 +450,7 @@ extern int nv_pat_mode;
         flush_cache_all();
     }
 #define CACHE_FLUSH()            nv_flush_cache_cpu(NULL)
-#define CACHE_FLUSH_ALL()        NV_ON_EACH_CPU(nv_flush_cache_cpu, NULL)
+#define CACHE_FLUSH_ALL()        on_each_cpu(nv_flush_cache_cpu, NULL, 1)
 #define WRITE_COMBINE_FLUSH()    mb()
 #endif
 #elif defined(NVCPU_PPC64LE)
@@ -520,8 +477,6 @@ extern int nv_pat_mode;
 
 #define NV_MAX_RECURRING_WARNING_MESSAGES 10
 
-#ifndef NVWATCH
-
 /* various memory tracking/debugging techniques
  * disabled for retail builds, enabled for debug builds
  */
@@ -695,62 +650,6 @@ static inline void nv_vunmap(NvUPtr vaddr, NvU32 page_count)
     NV_MEMDBG_REMOVE((void *)vaddr, page_count * PAGE_SIZE);
 }
 
-#endif /* !defined NVWATCH */
-
-#if defined(NV_SMP_CALL_FUNCTION_PRESENT)
-#if (NV_SMP_CALL_FUNCTION_ARGUMENT_COUNT == 4)
-#define NV_SMP_CALL_FUNCTION(func, info)                     \
-    ({                                                       \
-        int __ret = smp_call_function(func, info, 1, 1);     \
-        __ret;                                               \
-     })
-#elif (NV_SMP_CALL_FUNCTION_ARGUMENT_COUNT == 3)
-#define NV_SMP_CALL_FUNCTION(func, info)                     \
-    ({                                                       \
-        int __ret = smp_call_function(func, info, 1);        \
-        __ret;                                               \
-     })
-#else
-#error "NV_SMP_CALL_FUNCTION_ARGUMENT_COUNT value unrecognized!"
-#endif
-#elif defined(CONFIG_SMP)
-#error "NV_SMP_CALL_FUNCTION() undefined (smp_call_function() unavailable)!"
-#else
-// Non-SMP case:
-#define NV_SMP_CALL_FUNCTION(func, info)                     \
-    ({                                                       \
-        func(info);                                          \
-        0;                                                   \
-     })
-#endif
-
-#if defined(NV_ON_EACH_CPU_PRESENT)
-#if (NV_ON_EACH_CPU_ARGUMENT_COUNT == 4)
-#define NV_ON_EACH_CPU(func, info)                     \
-    ({                                                 \
-        int __ret = on_each_cpu(func, info, 1, 1);     \
-        __ret;                                         \
-     })
-#elif (NV_ON_EACH_CPU_ARGUMENT_COUNT == 3)
-#define NV_ON_EACH_CPU(func, info)                     \
-    ({                                                 \
-        int __ret = on_each_cpu(func, info, 1);        \
-        __ret;                                         \
-     })
-#else
-#error "NV_ON_EACH_CPU_ARGUMENT_COUNT value unrecognized!"
-#endif
-#elif defined(CONFIG_SMP)
-#error "NV_ON_EACH_CPU() undefined (on_each_cpu() unavailable)!"
-#else
-// Non-SMP case:
-#define NV_ON_EACH_CPU(func, info)                     \
-    ({                                                 \
-        func(info);                                    \
-        0;                                             \
-     })
-#endif
-
 #if defined(NV_GET_NUM_PHYSPAGES_PRESENT)
 #define NV_NUM_PHYSPAGES                get_num_physpages()
 #else
@@ -764,7 +663,7 @@ static inline void nv_vunmap(NvUPtr vaddr, NvU32 page_count)
 #define NV_COPY_FROM_USER(to, from, n)  copy_from_user(to, from, n)
 
 #define NV_IS_SUSER()                   capable(CAP_SYS_ADMIN)
-#define NV_PCI_DEVICE_NAME(dev)         ((dev)->pretty_name)
+#define NV_PCI_DEVICE_NAME(pci_dev)     ((pci_dev)->pretty_name)
 #define NV_CLI()                        local_irq_disable()
 #define NV_SAVE_FLAGS(eflags)           local_save_flags(eflags)
 #define NV_RESTORE_FLAGS(eflags)        local_irq_restore(eflags)
@@ -775,10 +674,10 @@ static inline void nv_vunmap(NvUPtr vaddr, NvU32 page_count)
 
 #define NV_NUM_CPUS()                   num_possible_cpus()
 
-static inline dma_addr_t nv_phys_to_dma(struct pci_dev *dev, NvU64 pa)
+static inline dma_addr_t nv_phys_to_dma(struct device *dev, NvU64 pa)
 {
 #if defined(NV_PHYS_TO_DMA_PRESENT)
-    return phys_to_dma(&dev->dev, pa);
+    return phys_to_dma(dev, pa);
 #elif defined(NV_XEN_SUPPORT_FULLY_VIRTUALIZED_KERNEL)
     return phys_to_machine(pa);
 #else
@@ -796,51 +695,41 @@ static inline dma_addr_t nv_phys_to_dma(struct pci_dev *dev, NvU64 pa)
 #define NV_DEVICE_MINOR_NUMBER(x)     minor((x)->i_rdev)
 #define NV_CONTROL_DEVICE_MINOR       255
 
-#define NV_PCI_DISABLE_DEVICE(dev)                            \
-    {                                                         \
-        NvU16 __cmd[2];                                       \
-        pci_read_config_word((dev), PCI_COMMAND, &__cmd[0]);  \
-        pci_disable_device(dev);                              \
-        pci_read_config_word((dev), PCI_COMMAND, &__cmd[1]);  \
-        __cmd[1] |= PCI_COMMAND_MEMORY;                       \
-        pci_write_config_word((dev), PCI_COMMAND,             \
-                (__cmd[1] | (__cmd[0] & PCI_COMMAND_IO)));    \
+#define NV_PCI_DISABLE_DEVICE(pci_dev)                           \
+    {                                                            \
+        NvU16 __cmd[2];                                          \
+        pci_read_config_word((pci_dev), PCI_COMMAND, &__cmd[0]); \
+        pci_disable_device(pci_dev);                             \
+        pci_read_config_word((pci_dev), PCI_COMMAND, &__cmd[1]); \
+        __cmd[1] |= PCI_COMMAND_MEMORY;                          \
+        pci_write_config_word((pci_dev), PCI_COMMAND,            \
+                (__cmd[1] | (__cmd[0] & PCI_COMMAND_IO)));       \
     }
 
-#define NV_PCI_RESOURCE_START(dev, bar) pci_resource_start(dev, (bar))
-#define NV_PCI_RESOURCE_SIZE(dev, bar)  pci_resource_len(dev, (bar))
-#define NV_PCI_RESOURCE_FLAGS(dev, bar) pci_resource_flags(dev, (bar))
+#define NV_PCI_RESOURCE_START(pci_dev, bar) pci_resource_start(pci_dev, (bar))
+#define NV_PCI_RESOURCE_SIZE(pci_dev, bar)  pci_resource_len(pci_dev, (bar))
+#define NV_PCI_RESOURCE_FLAGS(pci_dev, bar) pci_resource_flags(pci_dev, (bar))
 
 #if defined(NVCPU_X86)
-#define NV_PCI_RESOURCE_VALID(dev, bar)                                      \
-    ((NV_PCI_RESOURCE_START(dev, bar) != 0) &&                               \
-     (NV_PCI_RESOURCE_SIZE(dev, bar) != 0) &&                                \
-     (!((NV_PCI_RESOURCE_FLAGS(dev, bar) & PCI_BASE_ADDRESS_MEM_TYPE_64) &&  \
-       ((NV_PCI_RESOURCE_START(dev, bar) >> PAGE_SHIFT) > 0xfffffULL))))
+#define NV_PCI_RESOURCE_VALID(pci_dev, bar)                                     \
+    ((NV_PCI_RESOURCE_START(pci_dev, bar) != 0) &&                              \
+     (NV_PCI_RESOURCE_SIZE(pci_dev, bar) != 0) &&                               \
+     (!((NV_PCI_RESOURCE_FLAGS(pci_dev, bar) & PCI_BASE_ADDRESS_MEM_TYPE_64) && \
+       ((NV_PCI_RESOURCE_START(pci_dev, bar) >> PAGE_SHIFT) > 0xfffffULL))))
 #else
-#define NV_PCI_RESOURCE_VALID(dev, bar)                                      \
-    ((NV_PCI_RESOURCE_START(dev, bar) != 0) &&                               \
-     (NV_PCI_RESOURCE_SIZE(dev, bar) != 0))
+#define NV_PCI_RESOURCE_VALID(pci_dev, bar)                                     \
+    ((NV_PCI_RESOURCE_START(pci_dev, bar) != 0) &&                              \
+     (NV_PCI_RESOURCE_SIZE(pci_dev, bar) != 0))
 #endif
 
-#if defined(NV_PCI_DOMAIN_NR_PRESENT)
-#define NV_PCI_DOMAIN_NUMBER(dev)     (NvU32)pci_domain_nr(dev->bus)
-#else
-#define NV_PCI_DOMAIN_NUMBER(dev)     (0)
-#endif
-#define NV_PCI_BUS_NUMBER(dev)        (dev)->bus->number
-#define NV_PCI_DEVFN(dev)             (dev)->devfn
-#define NV_PCI_SLOT_NUMBER(dev)       PCI_SLOT(NV_PCI_DEVFN(dev))
+#define NV_PCI_DOMAIN_NUMBER(pci_dev) (NvU32)pci_domain_nr(pci_dev->bus)
+#define NV_PCI_BUS_NUMBER(pci_dev)    (pci_dev)->bus->number
+#define NV_PCI_DEVFN(pci_dev)         (pci_dev)->devfn
+#define NV_PCI_SLOT_NUMBER(pci_dev)   PCI_SLOT(NV_PCI_DEVFN(pci_dev))
 
-#if defined(NV_PCI_GET_CLASS_PRESENT)
-#define NV_PCI_DEV_PUT(dev)                    pci_dev_put(dev)
+#define NV_PCI_DEV_PUT(pci_dev)                pci_dev_put(pci_dev)
 #define NV_PCI_GET_DEVICE(vendor,device,from)  pci_get_device(vendor,device,from)
 #define NV_PCI_GET_CLASS(class,from)           pci_get_class(class,from)
-#else
-#define NV_PCI_DEV_PUT(dev)
-#define NV_PCI_GET_DEVICE(vendor,device,from)  pci_find_device(vendor,device,from)
-#define NV_PCI_GET_CLASS(class,from)           pci_find_class(class,from)
-#endif
 
 #if defined(CONFIG_X86_UV) && defined(NV_CONFIG_X86_UV)
 #define NV_GET_DOMAIN_BUS_AND_SLOT(domain,bus,devfn)                        \
@@ -900,7 +789,7 @@ static inline dma_addr_t nv_phys_to_dma(struct pci_dev *dev, NvU64 pa)
 #elif defined(NV_PCI_GET_DOMAIN_BUS_AND_SLOT_PRESENT)
 #define NV_GET_DOMAIN_BUS_AND_SLOT(domain,bus, devfn) \
     pci_get_domain_bus_and_slot(domain, bus, devfn)
-#elif defined(NV_PCI_GET_CLASS_PRESENT)
+#else
 #define NV_GET_DOMAIN_BUS_AND_SLOT(domain,bus,devfn)               \
    ({                                                              \
         struct pci_dev *__dev = NULL;                              \
@@ -916,14 +805,12 @@ static inline dma_addr_t nv_phys_to_dma(struct pci_dev *dev, NvU64 pa)
         }                                                          \
         __dev;                                                     \
     })
-#else
-#define NV_GET_DOMAIN_BUS_AND_SLOT(domain,bus,devfn) pci_find_slot(bus,devfn)
 #endif
 
 #if defined(NV_PCI_STOP_AND_REMOVE_BUS_DEVICE_PRESENT)  // introduced in 3.4.9
-#define NV_PCI_STOP_AND_REMOVE_BUS_DEVICE(dev) pci_stop_and_remove_bus_device(dev)
+#define NV_PCI_STOP_AND_REMOVE_BUS_DEVICE(pci_dev) pci_stop_and_remove_bus_device(pci_dev)
 #elif defined(NV_PCI_REMOVE_BUS_DEVICE_PRESENT) // introduced in 2.6
-#define NV_PCI_STOP_AND_REMOVE_BUS_DEVICE(dev) pci_remove_bus_device(dev)
+#define NV_PCI_STOP_AND_REMOVE_BUS_DEVICE(pci_dev) pci_remove_bus_device(pci_dev)
 #endif
 
 #define NV_PRINT_AT(nv_debug_level,at)                                           \
@@ -943,47 +830,6 @@ static inline dma_addr_t nv_phys_to_dma(struct pci_dev *dev, NvU64 pa)
             NV_VMA_OFFSET(vma), NV_VMA_PRIVATE(vma), NV_VMA_FILE(vma));                  \
     }
 
-/*
- * On Linux 2.6, we support both APM and ACPI power management. On Linux
- * 2.4, we support APM, only. ACPI support has been back-ported to the
- * Linux 2.4 kernel, but the Linux 2.4 driver model is not sufficient for
- * full ACPI support: it may work with some systems, but not reliably
- * enough for us to officially support this configuration.
- *
- * We support two Linux kernel power managment interfaces: the original
- * pm_register()/pm_unregister() on Linux 2.4 and the device driver model
- * backed PCI driver power management callbacks introduced with Linux
- * 2.6.
- *
- * The code below determines which interface to support on this kernel
- * version, if any; if built for Linux 2.6, it will also determine if the
- * kernel comes with ACPI or APM power management support.
- */
-#if defined(CONFIG_PM)
-#define NV_PM_SUPPORT_DEVICE_DRIVER_MODEL
-#if (defined(CONFIG_APM) || defined(CONFIG_APM_MODULE)) && !defined(CONFIG_ACPI)
-#define NV_PM_SUPPORT_NEW_STYLE_APM
-#endif
-#endif
-
-/*
- * On Linux 2.6 kernels >= 2.6.11, the PCI subsystem provides a new
- * interface that allows PCI drivers to determine the correct power state
- * for a given system power state; our suspend/resume callbacks now use
- * this interface and operate on PCI power state defines.
- *
- * Define these new PCI power state #define's here for compatibility with
- * older Linux 2.6 kernels.
- */
-#if !defined(PCI_D0)
-#define PCI_D0 PM_SUSPEND_ON
-#define PCI_D3hot PM_SUSPEND_MEM
-#endif
-
-#if !defined(NV_PM_MESSAGE_T_PRESENT)
-typedef u32 pm_message_t;
-#endif
-
 #ifndef minor
 # define minor(x) MINOR(x)
 #endif
@@ -1064,16 +910,11 @@ typedef void irqreturn_t;
     #endif
 #endif
 
-#if defined(NV_VM_INSERT_PAGE_PRESENT)
-#define NV_VM_INSERT_PAGE(vma, addr, page) \
-    vm_insert_page(vma, addr, page)
-#endif
-
 static inline int nv_remap_page_range(struct vm_area_struct *vma,
     unsigned long virt_addr, NvU64 phys_addr, NvU64 size, pgprot_t prot)
 {
     int ret = -1;
-#if defined(NV_REMAP_PFN_RANGE_PRESENT)
+
 #if defined(NV_4K_PAGE_ISOLATION_PRESENT) && defined(NV_PROT_4K_PAGE_ISOLATION)
     if ((size == PAGE_SIZE) &&
         ((pgprot_val(prot) & NV_PROT_4K_PAGE_ISOLATION) != 0))
@@ -1096,13 +937,7 @@ static inline int nv_remap_page_range(struct vm_area_struct *vma,
         ret = remap_pfn_range(vma, virt_addr, (phys_addr >> PAGE_SHIFT), size,
             prot);
     }
-#else
-    /*
-     * remap_page_range() was replaced by remap_pfn_range() in 2.6.10, but we
-     * still support 2.6.9.
-     */
-    ret = remap_page_range(vma, virt_addr, phys_addr, size, prot);
-#endif /* defined(NV_REMAP_PFN_RANGE_PRESENT) */
+
     return ret;
 }
 
@@ -1120,29 +955,70 @@ static inline int nv_io_remap_page_range(struct vm_area_struct *vma,
     return ret;
 }
 
-#define NV_PAGE_MASK    (NvU64)(long)PAGE_MASK
-
 #ifndef NV_VMF_INSERT_PFN_PRESENT
 typedef int vm_fault_t;
 #endif
 
+static inline vm_fault_t nv_insert_pfn(struct vm_area_struct *vma,
+    NvU64 virt_addr, NvU64 pfn, NvU32 extra_prot)
+{
+    /*
+     * vm_insert_pfn{,_prot} replaced with vmf_insert_pfn{,_prot} in Linux 4.20
+     */
+#if defined(NV_VMF_INSERT_PFN_PROT_PRESENT)
+    return vmf_insert_pfn_prot(vma, virt_addr, pfn,
+             __pgprot(pgprot_val(vma->vm_page_prot) | extra_prot));
+#else
+    int ret = -EINVAL;
+    /*
+     * Only PPC64LE (NV_4K_PAGE_ISOLATION_PRESENT) requires extra_prot to be
+     * used when remapping.
+     *
+     * vm_insert_pfn_prot() was added in Linux 4.4, whereas POWER9 support
+     * was added in Linux 4.8.
+     *
+     * Rather than tampering with the vma to make use of extra_prot with
+     * vm_insert_pfn() on older kernels, for now, just fail in this case, as
+     * it's not expected to be used currently.
+     */
+#if defined(NV_VM_INSERT_PFN_PROT_PRESENT)
+    ret = vm_insert_pfn_prot(vma, virt_addr, pfn,
+        __pgprot(pgprot_val(vma->vm_page_prot) | extra_prot));
+#elif !defined(NV_4K_PAGE_ISOLATION_PRESENT)
+    ret = vm_insert_pfn(vma, virt_addr, pfn);
+#endif
+    switch (ret)
+    {
+        case 0:
+        case -EBUSY:
+            /*
+             * EBUSY indicates that another thread already handled
+             * the faulted range.
+             */
+            return VM_FAULT_NOPAGE;
+        case -ENOMEM:
+            return VM_FAULT_OOM;
+        default:
+            break;
+    }
+#endif /* defined(NV_VM_INSERT_PFN_PROT_PRESENT) */
+    return VM_FAULT_SIGBUS;
+}
+
+
+#define NV_PAGE_MASK    (NvU64)(long)PAGE_MASK
+
 extern void *nvidia_stack_t_cache;
 
-// Changed in 2.6.23 via commit 20c2df83d25c6a95affe6157a4c9cac4cf5ffaac
-#if (NV_KMEM_CACHE_CREATE_ARGUMENT_COUNT == 5)
-#define NV_KMEM_CACHE_CREATE_FULL(name, size, align, flags, ctor) \
-    kmem_cache_create(name, size, align, flags, ctor)
-
-#else
-#define NV_KMEM_CACHE_CREATE_FULL(name, size, align, flags, ctor) \
-    kmem_cache_create(name, size, align, flags, ctor, NULL)
-#endif
-
 #define NV_KMEM_CACHE_CREATE(name, type)    \
-    NV_KMEM_CACHE_CREATE_FULL(name, sizeof(type), 0, 0, NULL)
+    kmem_cache_create(name, sizeof(type), 0, 0, NULL)
 
+/* The NULL pointer check is required for kernels older than 4.3 */
 #define NV_KMEM_CACHE_DESTROY(kmem_cache)   \
-    kmem_cache_destroy(kmem_cache)
+    if (kmem_cache != NULL)                 \
+    {                                       \
+        kmem_cache_destroy(kmem_cache);     \
+    }
 
 #define NV_KMEM_CACHE_ALLOC(kmem_cache)     \
     kmem_cache_alloc(kmem_cache, GFP_KERNEL)
@@ -1166,7 +1042,10 @@ static inline int nv_kmem_cache_alloc_stack(nvidia_stack_t **stack)
 static inline void nv_kmem_cache_free_stack(nvidia_stack_t *stack)
 {
 #if defined(NVCPU_X86) || defined(NVCPU_X86_64)
-    NV_KMEM_CACHE_FREE(stack, nvidia_stack_t_cache);
+    if (stack != NULL)
+    {
+        NV_KMEM_CACHE_FREE(stack, nvidia_stack_t_cache);
+    }
 #endif
 }
 
@@ -1200,7 +1079,7 @@ typedef struct nvidia_pte_s {
 
 typedef struct nv_alloc_s {
     struct nv_alloc_s *next;
-    struct pci_dev    *dev;
+    struct device     *dev;
     atomic_t       usage_count;
     unsigned int   flags;
     unsigned int   num_pages;
@@ -1282,7 +1161,8 @@ static inline NvBool nv_is_dma_direct(struct device *dev)
  * device.  So if SWIOTLB is enabled, we should avoid making
  * mapping calls.
  */
-static inline NvBool nv_dma_maps_swiotlb(struct pci_dev *dev)
+static inline NvBool
+nv_dma_maps_swiotlb(struct device *dev)
 {
     NvBool swiotlb_in_use = NV_FALSE;
 #if defined(CONFIG_SWIOTLB)
@@ -1302,9 +1182,9 @@ static inline NvBool nv_dma_maps_swiotlb(struct pci_dev *dev)
      * happen with the (NV_IS_EXPORT_SYMBOL_PRESENT_swiotlb_map_sg_attrs == 1)
      * case.
      */
-    const struct dma_map_ops *ops __attribute__ ((unused)) = get_dma_ops(&dev->dev);
+    const struct dma_map_ops *ops __attribute__ ((unused)) = get_dma_ops(dev);
       #else
-    const struct dma_mapping_ops *ops __attribute__ ((unused)) = get_dma_ops(&dev->dev);
+    const struct dma_mapping_ops *ops __attribute__ ((unused)) = get_dma_ops(dev);
       #endif
     #else
     const struct dma_mapping_ops *ops __attribute__ ((unused)) = dma_ops;
@@ -1387,11 +1267,7 @@ static inline NvBool nv_dma_maps_swiotlb(struct pci_dev *dev)
 typedef struct nv_dma_submap_s {
     NvU32 page_count;
     NvU32 sg_map_count;
-#if defined(NV_SG_TABLE_PRESENT)
     struct sg_table sgt;
-#else
-    struct scatterlist *sgl;
-#endif
 } nv_dma_submap_t;
 
 typedef struct nv_dma_map_s {
@@ -1413,7 +1289,7 @@ typedef struct nv_dma_map_s {
         } contig;
     } mapping;
 
-    struct pci_dev *dev;
+    struct pci_dev *pci_dev;
 } nv_dma_map_t;
 
 #define NV_FOR_EACH_DMA_SUBMAP(dm, sm, i)                                     \
@@ -1424,25 +1300,6 @@ typedef struct nv_dma_map_s {
 #define NV_DMA_SUBMAP_MAX_PAGES           ((NvU32)(NV_U32_MAX >> PAGE_SHIFT))
 #define NV_DMA_SUBMAP_IDX_TO_PAGE_IDX(s)  (s * NV_DMA_SUBMAP_MAX_PAGES)
 
-#if defined(NV_SG_TABLE_PRESENT)
-#define NV_DMA_SUBMAP_SCATTERLIST(sm)           sm->sgt.sgl
-#define NV_DMA_SUBMAP_SCATTERLIST_LENGTH(sm)    sm->sgt.orig_nents
-#else
-#define NV_DMA_SUBMAP_SCATTERLIST(sm)           sm->sgl
-#define NV_DMA_SUBMAP_SCATTERLIST_LENGTH(sm)    sm->page_count
-#endif
-
-#if defined(for_each_sg)
-    #define NV_FOR_EACH_DMA_SUBMAP_SG(sm, sg, i)                              \
-        for_each_sg(NV_DMA_SUBMAP_SCATTERLIST((sm)), (sg),                    \
-            (sm)->sg_map_count, (i))
-#else
-    #define NV_FOR_EACH_DMA_SUBMAP_SG(sm, sg, i)                              \
-        for ((i) = 0, (sg) = &NV_DMA_SUBMAP_SCATTERLIST(sm)[0];               \
-             (i) < (sm)->sg_map_count;                                        \
-             (sg) = &NV_DMA_SUBMAP_SCATTERLIST(sm)[++i])
-#endif
-
 /*
  * DO NOT use sg_alloc_table_from_pages on Xen Server, even if it's available.
  * This will glom multiple pages into a single sg element, which
@@ -1457,69 +1314,19 @@ typedef struct nv_dma_map_s {
             sm->page_count, 0,                                                \
             sm->page_count * PAGE_SIZE, NV_GFP_KERNEL) == 0) ? NV_OK :        \
                 NV_ERR_OPERATING_SYSTEM)
-
-    #define NV_FREE_DMA_SUBMAP_SCATTERLIST(sm)  sg_free_table(&sm->sgt)
-
-#else /* !defined(NV_SG_ALLOC_TABLE_FROM_PAGES_PRESENT) ||
-          defined(NV_DOM0_KERNEL_PRESENT) */
-    #if defined(NV_SG_TABLE_PRESENT)
-        #if defined(NV_SG_ALLOC_TABLE_PRESENT)
-            #define NV_ALLOC_DMA_SUBMAP_SCATTERLIST(dm, sm, i)                \
-                ((sg_alloc_table(&sm->sgt, sm->page_count, NV_GFP_KERNEL)) == \
-                    0 ? NV_OK : NV_ERR_OPERATING_SYSTEM)
-
-            #define NV_FREE_DMA_SUBMAP_SCATTERLIST(sm)  sg_free_table(&sm->sgt)
-
-        #else
-            #error "No known allocation function for sg_table present!"
-        #endif
-    #else /* !defined(NV_SG_TABLE_PRESENT) */
-        #define NV_ALLOC_DMA_SUBMAP_SCATTERLIST(dm, sm, i)                    \
-            os_alloc_mem((void **)&sm->sgl,                                    \
-                sm->page_count * sizeof(struct scatterlist))
-
-        #define NV_FREE_DMA_SUBMAP_SCATTERLIST(sm)  os_free_mem(sm->sgl)
-
-    #endif /* defined(NV_SG_TABLE_PRESENT) */
-
-#endif /* defined(NV_SG_ALLOC_TABLE_FROM_PAGES_PRESENT) &&
-          !defined(NV_DOM0_KERNEL_PRESENT) */
-
-typedef struct work_struct nv_task_t;
+#else
+    #define NV_ALLOC_DMA_SUBMAP_SCATTERLIST(dm, sm, i)                \
+        ((sg_alloc_table(&sm->sgt, sm->page_count, NV_GFP_KERNEL)) == \
+            0 ? NV_OK : NV_ERR_OPERATING_SYSTEM)
+#endif
 
 typedef struct nv_ibmnpu_info nv_ibmnpu_info_t;
 
 typedef struct nv_work_s {
-    nv_task_t task;
+    struct work_struct task;
     void *data;
 } nv_work_t;
 
-#define NV_WORKQUEUE_SCHEDULE(work) schedule_work(work)
-#define NV_WORKQUEUE_FLUSH()                           \
-    flush_scheduled_work();
-#if (NV_INIT_WORK_ARGUMENT_COUNT == 2)
-#define NV_WORKQUEUE_INIT(tq,handler,data)             \
-    {                                                  \
-        struct work_struct __work =                    \
-            __WORK_INITIALIZER(*(tq), handler);        \
-        *(tq) = __work;                                \
-    }
-#define NV_WORKQUEUE_DATA_T nv_task_t
-#define NV_WORKQUEUE_UNPACK_DATA(tq)                   \
-    container_of((tq), nv_work_t, task)
-#elif (NV_INIT_WORK_ARGUMENT_COUNT == 3)
-#define NV_WORKQUEUE_INIT(tq,handler,data)             \
-    {                                                  \
-        struct work_struct __work =                    \
-            __WORK_INITIALIZER(*(tq), handler, data);  \
-        *(tq) = __work;                                \
-    }
-#define NV_WORKQUEUE_DATA_T void
-#define NV_WORKQUEUE_UNPACK_DATA(tq) (nv_work_t *)(tq)
-#else
-#error "NV_INIT_WORK_ARGUMENT_COUNT value unrecognized!"
-#endif
-
 #define NV_MAX_REGISTRY_KEYS_LENGTH   512
 
 typedef enum
@@ -1528,18 +1335,26 @@ typedef enum
     NV_DEV_STACK_ISR,
     NV_DEV_STACK_ISR_BH,
     NV_DEV_STACK_ISR_BH_UNLOCKED,
-    NV_DEV_STACK_PCI_CFGCHK,
+    NV_DEV_STACK_GPU_WAKEUP,
     NV_DEV_STACK_COUNT
 } nvidia_linux_dev_stack_t;
 
+/* Linux version of the opaque type used for os_queue_work_item() */
+struct os_work_queue {
+    nv_kthread_q_t nvk;
+};
+
 /* linux-specific version of old nv_state_t */
 /* this is a general os-specific state structure. the first element *must* be
    the general state structure, for the generic unix-based code */
 typedef struct nv_linux_state_s {
     nv_state_t nv_state;
-    atomic_t usage_count;
 
-    struct pci_dev *dev;
+    atomic_t usage_count;
+    NvU32    suspend_count;
+
+    struct device  *dev;
+    struct pci_dev *pci_dev;
 
     /* IBM-NPU info associated with this GPU */
     nv_ibmnpu_info_t *npu;
@@ -1592,6 +1407,16 @@ typedef struct nv_linux_state_s {
         struct backlight_device *dev;
         NvU32 displayId;
     } backlight;
+
+    /* Per-GPU queue */
+    struct os_work_queue queue;
+
+    /* GPU user mapping revocation/remapping (only for non-CTL device) */
+    struct semaphore mmap_lock; /* Protects all fields in this category */
+    struct list_head open_files;
+    NvBool all_mappings_revoked;
+    NvBool safe_to_mmap;
+    NvBool gpu_wakeup_callback_needed;
 } nv_linux_state_t;
 
 extern nv_linux_state_t *nv_linux_devices;
@@ -1611,6 +1436,40 @@ extern struct semaphore nv_linux_devices_lock;
 #define LOCK_NV_LINUX_DEVICES()     down(&nv_linux_devices_lock)
 #define UNLOCK_NV_LINUX_DEVICES()   up(&nv_linux_devices_lock)
 
+/*
+ * Macros to synchronize system power management transitions,
+ * and to protect the global system PM state.  The procfs power
+ * management interface acquires this lock in write mode for
+ * the duration of the sleep operation, any other paths accessing
+ * device state must acquire the lock in read mode.
+ */
+extern struct rw_semaphore nv_system_pm_lock;
+
+/*
+ * Approximate what down_read_interruptible() would accomplish
+ * it were available.  This is currently only needed to catch
+ * the fake signals sent from the freezer in the kernel's system
+ * suspend/hibernate paths.
+ */
+static inline int nv_read_lock_system_pm_lock_interruptible(void)
+{
+    while (!down_read_trylock(&nv_system_pm_lock))
+    {
+        if (signal_pending(current))
+            return -EINTR;
+        cond_resched();
+    }
+    return 0;
+}
+
+#define NV_READ_LOCK_SYSTEM_PM_LOCK()   down_read(&nv_system_pm_lock)
+#define NV_READ_LOCK_SYSTEM_PM_LOCK_INTERRUPTIBLE() \
+    nv_read_lock_system_pm_lock_interruptible()
+#define NV_READ_UNLOCK_SYSTEM_PM_LOCK() up_read(&nv_system_pm_lock)
+
+#define NV_WRITE_LOCK_SYSTEM_PM_LOCK()   down_write(&nv_system_pm_lock)
+#define NV_WRITE_UNLOCK_SYSTEM_PM_LOCK() up_write(&nv_system_pm_lock)
+
 extern NvBool nv_ats_supported;
 
 #if defined(NV_LINUX_ACPI_EVENTS_SUPPORTED)
@@ -1676,12 +1535,16 @@ typedef struct
     int event_pending;
     nv_spinlock_t fp_lock;
     wait_queue_head_t waitqueue;
+    nv_kthread_q_item_t deferred_close_q_item;
     off_t off;
     NvU32 minor_num;
     NvHandle hExportedRmObject;
     NvU32 *attached_gpus;
     size_t num_attached_gpus;
     nv_alloc_mapping_context_t mmap_context;
+    struct address_space mapping;
+
+    struct list_head entry;
 } nv_file_private_t;
 
 #define NV_SET_FILE_PRIVATE(filep,data) ((filep)->private_data = (data))
@@ -1731,112 +1594,28 @@ typedef struct
 
 
 
-#if (NV_PCI_SAVE_STATE_ARGUMENT_COUNT == 2)
-/*
- * On Linux 2.6.9, pci_(save|restore)_state() requires a storage buffer to be
- * passed in by the caller. This went away in 2.6.10, so until we can drop
- * this support, use the original config space storage buffer for this call.
- */
-static inline void nv_pci_save_state(struct pci_dev *dev)
-{
-    nv_linux_state_t *nvl = pci_get_drvdata(dev);
-    nv_state_t *nv = NV_STATE_PTR(nvl);
-    pci_save_state(dev, &nv->pci_cfg_space[0]);
-}
-
-static inline void nv_pci_restore_state(struct pci_dev *dev)
-{
-    nv_linux_state_t *nvl = pci_get_drvdata(dev);
-    nv_state_t *nv = NV_STATE_PTR(nvl);
-    pci_restore_state(dev, &nv->pci_cfg_space[0]);
-}
-#else
-#define nv_pci_save_state(dev)      pci_save_state(dev)
-#define nv_pci_restore_state(dev)   pci_restore_state(dev)
-#endif
-
 #define NV_PCIE_CFG_MAX_OFFSET 0x1000
 
 #include "nv-proto.h"
 
-#if defined(NVCPU_PPC64LE)
 /*
- * It is OK to stub this for PowerPC platforms because the conditions
- * below with X servers does not apply to PPC.
- */
-static inline NV_STATUS nv_check_pci_config_space_and_gpu_state(
-                            nvidia_stack_t *sp, nv_state_t *nv, NvBool checkBar,
-                            NvBool acquireLocks, NvBool mayBlock)
-{
-    return NV_OK;
-}
-#else
-/*
- * Verify that access to PCI configuration space wasn't modified
- * by a third party.  Unfortunately, some X servers disable
- * memory access in PCI configuration space at various times (such
- * as when restoring initial PCI configuration space settings
- * during VT switches or when driving multiple GPUs), which may
- * cause BAR[n] reads/writes to fail and/or inhibit GPU initiator
- * functionality.
- * Also check if GPU is present on the bus by checking flag
+ * Check if GPU is present on the bus by checking flag
  * NV_FLAG_IN_SURPRISE_REMOVAL(set when eGPU is removed from TB3).
  */
-static inline NV_STATUS nv_check_pci_config_space_and_gpu_state(
-                            nvidia_stack_t *sp, nv_state_t *nv, NvBool checkBar,
-                            NvBool acquireLocks, NvBool mayBlock)
+static inline NV_STATUS nv_check_gpu_state(nv_state_t *nv)
 {
-
+#if !defined(NVCPU_PPC64LE)
     if (NV_IS_DEVICE_IN_SURPRISE_REMOVAL(nv))
     {
         return NV_ERR_GPU_IS_LOST;
     }
-
-    if (!NV_IS_GVI_DEVICE(nv) &&
-        (((nv)->flags & NV_FLAG_SKIP_CFG_CHECK) == 0) &&
-        (((nv)->flags & NV_FLAG_CONTROL) == 0))
-    {
-        nv_check_pci_config_space(nv, checkBar);
-    }
+#endif
 
     return NV_OK;
 }
-#endif
-
-extern int nv_update_memory_types;
 
 extern NvU32 NVreg_EnableUserNUMAManagement;
 
-#if defined(NVCPU_X86) || defined(NVCPU_X86_64)
-/*
- * On Linux/x86-64 (and recent Linux/x86) kernels, the PAGE_KERNEL
- * and PAGE_KERNEL_NOCACHE protection bit masks include _PAGE_NX
- * to indicate that the no-execute protection page feature is used
- * for the page in question.
- *
- * We need to be careful to mask out _PAGE_NX when the host system
- * doesn't support this feature or when it's disabled: the kernel
- * may not do this in its implementation of the change_page_attr()
- * interface.
- */
-#ifndef X86_FEATURE_NX
-#define X86_FEATURE_NX (1*32+20)
-#endif
-#ifndef boot_cpu_has
-#define boot_cpu_has(x) test_bit(x, boot_cpu_data.x86_capability)
-#endif
-#ifndef MSR_EFER
-#define MSR_EFER 0xc0000080
-#endif
-#ifndef EFER_NX
-#define EFER_NX (1 << 11)
-#endif
-#ifndef _PAGE_NX
-#define _PAGE_NX ((NvU64)1 << 63)
-#endif
-extern NvU64 __nv_supported_pte_mask;
-#endif
-
 extern nv_pci_info_t nv_assign_gpu_pci_info[NV_MAX_DEVICES];
 extern NvU32 nv_assign_gpu_count;
 
@@ -1883,11 +1662,12 @@ static inline NvU64 nv_pci_bus_address(nv_linux_state_t *nvl, NvU8 bar_index)
 {
     NvU64 bus_addr = 0;
 #if defined(NV_PCI_BUS_ADDRESS_PRESENT)
-    bus_addr = pci_bus_address(nvl->dev, bar_index);
+    bus_addr = pci_bus_address(nvl->pci_dev, bar_index);
 #elif defined(CONFIG_PCI)
     struct pci_bus_region region;
 
-    pcibios_resource_to_bus(nvl->dev, &region, &nvl->dev->resource[bar_index]);
+    pcibios_resource_to_bus(nvl->pci_dev, &region,
+                            &nvl->pci_dev->resource[bar_index]);
     bus_addr = region.start;
 #endif
     return bus_addr;
