@@ -1,6 +1,6 @@
 /* _NVRM_COPYRIGHT_BEGIN_
  *
- * Copyright 2014-2015 by NVIDIA Corporation.  All rights reserved.  All
+ * Copyright 2014-2019 by NVIDIA Corporation.  All rights reserved.  All
  * information contained herein is proprietary and confidential to NVIDIA
  * Corporation.  Any use, reproduction, or disclosure without the written
  * permission of NVIDIA Corporation is prohibited.
@@ -55,6 +55,9 @@
 // Output flags
 #define UVM_PMA_ALLOCATE_RESULT_IS_ZERO         NVBIT(0)
 
+// Input flags to pmaFree
+#define UVM_PMA_FREE_IS_ZERO                    NVBIT(0)
+
 //
 // Indicate that the PMA operation is being done from one of the PMA eviction
 // callbacks.
@@ -156,9 +159,12 @@ typedef struct UvmGpuChannelInstanceInfo_tag
     // Out: Whether the channel belongs to a TSG or not
     NvBool bTsgChannel;
 
-    // Out: ID of the tsg to which the channel belongs.
+    // Out: ID of the TSG to which the channel belongs
     NvU32 tsgId;
 
+    // Out: Maximum number of subcontexts in the TSG to which the channel belongs
+    NvU32 tsgMaxSubctxCount;
+
     // Out: Number of channel resources (local CTX buffers + global CTX buffers)
     // associated with the channel.
     NvU32 resourceCount;
@@ -181,6 +187,15 @@ typedef struct UvmGpuChannelInstanceInfo_tag
 
 
 
+
+
+
+
+
+
+
+
+
 } UvmGpuChannelInstanceInfo;
 
 typedef struct UvmGpuChannelResourceBindParams_tag
@@ -198,13 +213,15 @@ typedef struct UvmGpuChannelResourceBindParams_tag
     NvBool globalResource;
 } UvmGpuChannelResourceBindParams;
 
-typedef struct UvmGpuChannelPointers_tag
+typedef struct UvmGpuChannelInfo_tag
 {
     volatile unsigned *gpGet;
     volatile unsigned *gpPut;
     UvmGpuPointer     *gpFifoEntries;
     unsigned           numGpFifoEntries;
     unsigned           channelClassNum;
+
+    // The errorNotifier is filled out when the channel hits an RC error.
     NvNotification    *errorNotifier;
     NvU32              hwChannelId;
     volatile unsigned *dummyBar1Mapping;
@@ -214,7 +231,7 @@ typedef struct UvmGpuChannelPointers_tag
     // happens after CE allocation.
     volatile NvU32    *workSubmissionOffset;
     NvU32              workSubmissionToken;
-} UvmGpuChannelPointers;
+} UvmGpuChannelInfo;
 
 typedef enum
 {
@@ -233,6 +250,9 @@ typedef struct UvmGpuChannelAllocParams_tag
     // The next two fields store UVM_BUFFER_LOCATION values
     NvU32 gpFifoLoc;
     NvU32 gpPutLoc;
+
+    // Index of the engine the channel will be bound to
+    NvU32 engineIndex;
 } UvmGpuChannelAllocParams;
 
 // The max number of Copy Engines supported by a GPU.
@@ -288,7 +308,10 @@ typedef enum
     UVM_PEER_LINK_TYPE_NONE,
     UVM_PEER_LINK_TYPE_PCIE,
     UVM_PEER_LINK_TYPE_NVLINK_1,
-    UVM_PEER_LINK_TYPE_NVLINK_2
+    UVM_PEER_LINK_TYPE_NVLINK_2,
+
+
+
 } UVM_PEER_LINK_TYPE;
 
 typedef struct UvmGpuCaps_tag
@@ -496,6 +519,44 @@ typedef struct UvmPmaAllocationOptions_tag
     NvU32 resultFlags;          // valid if the allocation function returns NV_OK
 } UvmPmaAllocationOptions;
 
+/*******************************************************************************
+    uvmEventSuspend
+    This function will be called by the GPU driver to signal to UVM that the
+    system is about to enter a sleep state.  When it is called, the
+    following assumptions/guarantees are valid/made:
+
+      * User channels have been preempted and disabled
+      * UVM channels are still running normally and will continue to do
+        so until after this function returns control
+      * User threads are still running, but can no longer issue system
+        system calls to the GPU driver
+      * Until exit from this function, UVM is allowed to make full use of
+        the GPUs under its control, as well as of the GPU driver
+
+    Upon return from this function, UVM may not access GPUs under its control
+    until the GPU driver calls uvmEventResume().  It may still receive
+    calls to uvmEventIsrTopHalf() during this time, which it should return
+    NV_ERR_NO_INTR_PENDING from.  It will not receive any other calls.
+*/
+typedef NV_STATUS (*uvmEventSuspend_t) (void);
+
+/*******************************************************************************
+    uvmEventResume
+    This function will be called by the GPU driver to signal to UVM that the
+    system has exited a previously entered sleep state.  When it is called,
+    the following assumptions/guarantees are valid/made:
+
+      * UVM is again allowed to make full use of the GPUs under its
+        control, as well as of the GPU driver
+      * UVM channels are running normally
+      * User channels are still preempted and disabled
+      * User threads are again running, but still cannot issue system
+        calls to the GPU driver, nor submit new work
+
+    Upon return from this function, UVM is expected to be fully functional.
+*/
+typedef NV_STATUS (*uvmEventResume_t) (void);
+
 /*******************************************************************************
     uvmEventStartDevice
     This function will be called by the GPU driver once it has finished its
@@ -551,6 +612,8 @@ typedef void (*uvmEventIsrTopHalf_t) (void);
 
 struct UvmOpsUvmEvents
 {
+    uvmEventSuspend_t     suspend;
+    uvmEventResume_t      resume;
     uvmEventStartDevice_t startDevice;
     uvmEventStopDevice_t  stopDevice;
     uvmEventIsrTopHalf_t  isrTopHalf;
@@ -654,7 +717,7 @@ typedef struct UvmGpuAccessCntrConfig_tag
     NvU32 threshold;
 } UvmGpuAccessCntrConfig;
 
-typedef UvmGpuChannelPointers gpuChannelInfo;
+typedef UvmGpuChannelInfo gpuChannelInfo;
 typedef UvmGpuChannelAllocParams gpuChannelAllocParams;
 typedef UvmGpuCaps gpuCaps;
 typedef UvmGpuCopyEngineCaps gpuCeCaps;
