@@ -1,6 +1,6 @@
  /***************************************************************************\
 |*                                                                           *|
-|*       Copyright 1993-2018 NVIDIA, Corporation.  All rights reserved.      *|
+|*       Copyright 1993-2019 NVIDIA, Corporation.  All rights reserved.      *|
 |*                                                                           *|
 |*     NOTICE TO USER:   The source code  is copyrighted under  U.S. and     *|
 |*     international laws.  NVIDIA, Corp. of Sunnyvale,  California owns     *|
@@ -14,7 +14,7 @@
 |*     tion and  internal comments to the code,  notices to the end user     *|
 |*     as follows:                                                           *|
 |*                                                                           *|
-|*     Copyright  1993-2015  NVIDIA,  Corporation.   NVIDIA  has  design     *|
+|*     Copyright  1993-2019  NVIDIA,  Corporation.   NVIDIA  has  design     *|
 |*     patents and patents pending in the U.S. and foreign countries.        *|
 |*                                                                           *|
 |*     NVIDIA, CORPORATION MAKES NO REPRESENTATION ABOUT THE SUITABILITY     *|
@@ -55,10 +55,14 @@ extern "C" {
 //
 // STUPID HACK FOR CL 19434692.  Will revert when fix CL is delivered bfm -> chips_a.
 #ifndef BIT
-#define BIT(b)                  (1<<(b))
+#define BIT(b)                  (1U<<(b))
+#endif
+#ifndef BIT32
+#define BIT32(b)                ((NvU32)1U<<(b))
+#endif
+#ifndef BIT64
+#define BIT64(b)                ((NvU64)1U<<(b))
 #endif
-#define BIT32(b)                ((NvU32)1<<(b))
-#define BIT64(b)                ((NvU64)1<<(b))
 
 #endif
 
@@ -66,9 +70,18 @@ extern "C" {
 // It is recommended to use the following bit macros to avoid macro name
 // collisions with other src code bases.
 //
-#define NVBIT(b)                  (1<<(b))
-#define NVBIT32(b)                ((NvU32)1<<(b))
-#define NVBIT64(b)                ((NvU64)1<<(b))
+#ifndef NVBIT
+#define NVBIT(b)                  (1U<<(b))
+#endif
+#ifndef NVBIT_TYPE
+#define NVBIT_TYPE(b, t)          (((t)1U)<<(b))
+#endif
+#ifndef NVBIT32
+#define NVBIT32(b)                NVBIT_TYPE(b, NvU32)
+#endif
+#ifndef NVBIT64
+#define NVBIT64(b)                NVBIT_TYPE(b, NvU64)
+#endif
 
 // Index of the 'on' bit (assuming that there is only one).
 // Even if multiple bits are 'on', result is in range of 0-31.
@@ -122,7 +135,7 @@ extern "C" {
 #define DRF_EXTENT(drf)         (NV_TRUE?drf)  // much better
 #define DRF_SHIFT(drf)          (((NvU32)DRF_BASE(drf)) % 32U)
 #define DRF_SHIFT_RT(drf)       (((NvU32)DRF_EXTENT(drf)) % 32U)
-#define DRF_MASK(drf)           (0xFFFFFFFFU>>(31U-(((NvU32)DRF_EXTENT(drf)) % 32U)+(((NvU32)DRF_BASE(drf)) % 32U)))
+#define DRF_MASK(drf)           (0xFFFFFFFFU>>(31U - DRF_SHIFT_RT(drf) + DRF_SHIFT(drf)))
 #endif
 #define DRF_DEF(d,r,f,c)        (((NvU32)(NV ## d ## r ## f ## c))<<DRF_SHIFT(NV ## d ## r ## f))
 #define DRF_NUM(d,r,f,n)        ((((NvU32)(n))&DRF_MASK(NV ## d ## r ## f))<<DRF_SHIFT(NV ## d ## r ## f))
@@ -198,6 +211,23 @@ extern "C" {
 // 32 Bit Versions
 //
 
+#ifdef NV_MISRA_COMPLIANCE_REQUIRED
+#define FLD_SET_DRF(d,r,f,c,v)                  (((NvU32)(v) & ~DRF_SHIFTMASK(NV##d##r##f)) | DRF_DEF(d,r,f,c))
+#define FLD_SET_DRF_NUM(d,r,f,n,v)              (((NvU32)(v) & ~DRF_SHIFTMASK(NV##d##r##f)) | DRF_NUM(d,r,f,n))
+// FLD_SET_DRF_DEF is deprecated! Use an explicit assignment with FLD_SET_DRF instead.
+#define FLD_SET_DRF_DEF(d,r,f,c,v)              ((v) = ((NvU32)(v) & ~DRF_SHIFTMASK(NV##d##r##f)) | DRF_DEF(d,r,f,c))
+#define FLD_IDX_SET_DRF(d,r,f,i,c,v)            (((NvU32)(v) & ~DRF_SHIFTMASK(NV##d##r##f(i))) | DRF_IDX_DEF(d,r,f,i,c))
+#define FLD_IDX_OFFSET_SET_DRF(d,r,f,i,o,c,v)   (((NvU32)(v) & ~DRF_SHIFTMASK(NV##d##r##f(i,o))) | DRF_IDX_OFFSET_DEF(d,r,f,i,o,c))
+#define FLD_IDX_SET_DRF_DEF(d,r,f,i,c,v)        (((NvU32)(v) & ~DRF_SHIFTMASK(NV##d##r##f(i))) | DRF_IDX_DEF(d,r,f,i,c))
+#define FLD_IDX_SET_DRF_NUM(d,r,f,i,n,v)        (((NvU32)(v) & ~DRF_SHIFTMASK(NV##d##r##f(i))) | DRF_IDX_NUM(d,r,f,i,n))
+#define FLD_SET_DRF_IDX(d,r,f,c,i,v)            (((NvU32)(v) & ~DRF_SHIFTMASK(NV##d##r##f)) | DRF_DEF(d,r,f,c(i)))
+
+#define FLD_TEST_DRF(d,r,f,c,v)                 ((DRF_VAL(d, r, f, (v)) == (NvU32)(NV##d##r##f##c)))
+#define FLD_TEST_DRF_AND(d,r,f,c,v)             ((DRF_VAL(d, r, f, (v)) & (NvU32)(NV##d##r##f##c)) != 0U)
+#define FLD_TEST_DRF_NUM(d,r,f,n,v)             ((DRF_VAL(d, r, f, (v)) == (NvU32)(n)))
+#define FLD_IDX_TEST_DRF(d,r,f,i,c,v)           ((DRF_IDX_VAL(d, r, f, i, (v)) == (NvU32)(NV##d##r##f##c)))
+#define FLD_IDX_OFFSET_TEST_DRF(d,r,f,i,o,c,v)  ((DRF_IDX_OFFSET_VAL(d, r, f, i, o, (v)) == (NvU32)(NV##d##r##f##c)))
+#else
 #define FLD_SET_DRF(d,r,f,c,v)                  (((v) & ~DRF_SHIFTMASK(NV##d##r##f)) | DRF_DEF(d,r,f,c))
 #define FLD_SET_DRF_NUM(d,r,f,n,v)              (((v) & ~DRF_SHIFTMASK(NV##d##r##f)) | DRF_NUM(d,r,f,n))
 // FLD_SET_DRF_DEF is deprecated! Use an explicit assignment with FLD_SET_DRF instead.
@@ -213,6 +243,7 @@ extern "C" {
 #define FLD_TEST_DRF_NUM(d,r,f,n,v)             ((DRF_VAL(d, r, f, (v)) == (n)))
 #define FLD_IDX_TEST_DRF(d,r,f,i,c,v)           ((DRF_IDX_VAL(d, r, f, i, (v)) == NV##d##r##f##c))
 #define FLD_IDX_OFFSET_TEST_DRF(d,r,f,i,o,c,v)  ((DRF_IDX_OFFSET_VAL(d, r, f, i, o, (v)) == NV##d##r##f##c))
+#endif
 
 #define REF_DEF(drf,d)            (((drf ## d)&DRF_MASK(drf))<<DRF_SHIFT(drf))
 #define REF_VAL(drf,v)            (((v)>>DRF_SHIFT(drf))&DRF_MASK(drf))
@@ -663,11 +694,11 @@ nvPrevPow2_U32(const NvU32 x )
 #endif
 
 #ifndef NV_IS_ALIGNED
-#define NV_IS_ALIGNED(v, gran)      (0 == ((v) & ((gran) - 1)))
+#define NV_IS_ALIGNED(v, gran)      (0U == ((v) & ((gran) - 1U)))
 #endif
 
 #ifndef NV_IS_ALIGNED64
-#define NV_IS_ALIGNED64(v, gran)      (0 == ((v) & (((NvU64)gran) - 1)))
+#define NV_IS_ALIGNED64(v, gran)      (0U == ((v) & (((NvU64)gran) - 1U)))
 #endif
 
 #ifndef NVMISC_MEMSET
