@@ -1,5 +1,5 @@
 /*******************************************************************************
-    Copyright (c) 2015 NVIDIA Corporation
+    Copyright (c) 2015-2019 NVIDIA Corporation
 
     Permission is hereby granted, free of charge, to any person obtaining a copy
     of this software and associated documentation files (the "Software"), to
@@ -93,6 +93,26 @@ module_param(uvm_perf_fault_max_throttle_per_service, uint, S_IRUGO);
 static unsigned uvm_perf_fault_coalesce = 1;
 module_param(uvm_perf_fault_coalesce, uint, S_IRUGO);
 
+// This function is used for both the initial fault buffer initialization and
+// the power management resume path.
+static void fault_buffer_reinit_replayable_faults(uvm_gpu_t *gpu)
+{
+    uvm_replayable_fault_buffer_info_t *replayable_faults = &gpu->fault_buffer_info.replayable;
+
+    // Read the current get/put pointers, as this might not be the first time
+    // we take control of the fault buffer since the GPU was initialized,
+    // or since we may need to bring UVM's cached copies back in sync following
+    // a sleep cycle.
+    replayable_faults->cached_get = gpu->fault_buffer_hal->read_get(gpu);
+    replayable_faults->cached_put = gpu->fault_buffer_hal->read_put(gpu);
+
+    // (Re-)enable fault prefetching
+    if (gpu->fault_buffer_info.prefetch_faults_enabled)
+        gpu->arch_hal->enable_prefetch_faults(gpu);
+    else
+        gpu->arch_hal->disable_prefetch_faults(gpu);
+}
+
 // There is no error handling in this function. The caller is in charge of
 // calling fault_buffer_deinit_replayable_faults on failure.
 static NV_STATUS fault_buffer_init_replayable_faults(uvm_gpu_t *gpu)
@@ -150,11 +170,6 @@ static NV_STATUS fault_buffer_init_replayable_faults(uvm_gpu_t *gpu)
         return status;
     }
 
-    // Read current get/put pointers as this might not be the first time we have taken control of the fault buffer
-    // since the GPU was initialized
-    replayable_faults->cached_get = gpu->fault_buffer_hal->read_get(gpu);
-    replayable_faults->cached_put = gpu->fault_buffer_hal->read_put(gpu);
-
     replayable_faults->replay_policy = uvm_perf_fault_replay_policy < UVM_PERF_FAULT_REPLAY_POLICY_MAX?
                                            uvm_perf_fault_replay_policy:
                                            UVM_PERF_FAULT_REPLAY_POLICY_DEFAULT;
@@ -171,14 +186,9 @@ static NV_STATUS fault_buffer_init_replayable_faults(uvm_gpu_t *gpu)
     }
 
     // Re-enable fault prefetching just in case it was disabled in a previous run
-    if (gpu->prefetch_fault_supported) {
-        gpu->arch_hal->enable_prefetch_faults(gpu);
-        gpu->fault_buffer_info.prefetch_faults_enabled = true;
-    }
-    else {
-        gpu->arch_hal->disable_prefetch_faults(gpu);
-        gpu->fault_buffer_info.prefetch_faults_enabled = false;
-    }
+    gpu->fault_buffer_info.prefetch_faults_enabled = gpu->prefetch_fault_supported;
+
+    fault_buffer_reinit_replayable_faults(gpu);
 
     return NV_OK;
 }
@@ -248,6 +258,15 @@ fail:
     return status;
 }
 
+// Reinitialize state relevant to replayable fault handling after returning
+// from a power management cycle.
+void uvm_gpu_fault_buffer_resume(uvm_gpu_t *gpu)
+{
+    UVM_ASSERT(gpu->replayable_faults_supported);
+
+    fault_buffer_reinit_replayable_faults(gpu);
+}
+
 void uvm_gpu_fault_buffer_deinit(uvm_gpu_t *gpu)
 {
     NV_STATUS status = NV_OK;
@@ -2220,3 +2239,41 @@ NV_STATUS uvm8_test_set_prefetch_faults_reenable_lapse(UVM_TEST_SET_PREFETCH_FAU
 
     return NV_OK;
 }
+
+NV_STATUS uvm8_test_drain_replayable_faults(UVM_TEST_DRAIN_REPLAYABLE_FAULTS_PARAMS *params, struct file *filp)
+{
+    uvm_gpu_t *gpu;
+    NV_STATUS status = NV_OK;
+    uvm_spin_loop_t spin;
+    bool pending = true;
+    uvm_va_space_t *va_space = uvm_va_space_get(filp);
+
+    gpu = uvm_va_space_retain_gpu_by_uuid(va_space, &params->gpu_uuid);
+    if (!gpu)
+        return NV_ERR_INVALID_DEVICE;
+
+    uvm_spin_loop_init(&spin);
+
+    do {
+        uvm_gpu_replayable_faults_isr_lock(gpu);
+        pending = uvm_gpu_replayable_faults_pending(gpu);
+        uvm_gpu_replayable_faults_isr_unlock(gpu);
+
+        if (!pending)
+            break;
+
+        if (fatal_signal_pending(current)) {
+            status = NV_ERR_SIGNAL_PENDING;
+            break;
+        }
+
+        UVM_SPIN_LOOP(&spin);
+    } while (uvm_spin_loop_elapsed(&spin) < params->timeout_ns);
+
+    if (pending && status == NV_OK)
+        status = NV_ERR_TIMEOUT;
+
+    uvm_gpu_release(gpu);
+
+    return status;
+}
